﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using MLogics.VisXpert.Online.ClientV2;//This is the Namespace where most of your Variable values functionality lives
using MLogics.VisXpert.Online;

namespace Example_Get_or_Set_Variables_V2
{
    public partial class GetSetVarsDialog : Form
    {
        /// <summary>
        /// this is the Application Interface V2. It represents an connection to the Communication Module and
        /// allows you to exchange data with the Communication module. 
        /// It has Facilities to Manipulate Variable Values and to Subscribe to Events and Value changes
        /// 
        /// Usually you will have one instance of this Per Application. 
        /// To facilitate this, you can also use the ApplicationIntf.Instance, which is an Singleton
        /// 
        /// The V2 has the benefit of being the future Application interface and being about 2x faster than V1
        /// Porting from V1 to V2 is relatively easy, but not a Drop in replacement, as the structure of both interfaces is different
        /// an important distinction is that ValueItems are now called ClientVariables and the Client does not manage
        /// the Connection to the Communication module. 
        /// the connection is managed by the Module Connection, which can be used for more than the Client
        /// </summary>
        private ModuleConnection ModCon;
        private Client AppIntf;

        public GetSetVarsDialog()
        {
            //First we construct an Module Connection. This connection gives us access to basic functionality
            //and also tells us when the Communication module wants us to shut down.
            ModCon = new ModuleConnection("MeasurementView 1");

            InitializeComponent();
            RefreshConnectionStateView();
            RefreshSubscribedView();
        }

    private void RefreshConnectionStateView()
        {
            groupBoxReadWrite.Enabled = ModCon.isConnected;
            groupBoxSubscribe.Enabled = ModCon.isConnected;
            buttonDisconnect.Enabled = ModCon.isConnected;
            buttonConnect.Enabled = !ModCon.isConnected;

            if (ModCon.isConnected) labelConnectionState.Text = "Connected";
            else labelConnectionState.Text = "Disconnected";
        }

        /// <summary>
        /// We need to hold on tho an Handle of the ValueItem we subscribed to. 
        /// This way we know if we have an subscribed item or not, but we also keep the item 
        /// Alive, so it does not get Garbage collected
        /// </summary>
        ClientVariable SubscribedVariable;

        private void RefreshSubscribedView()
        {
            buttonSubscribe.Enabled = SubscribedVariable == null;
            buttonUnsubscribe.Enabled = SubscribedVariable != null;
        }

        /// <summary>
        /// This represents our generic error handler. Of course you should differentiate exception by type and
        /// Recover where possible, but in our demonstration we only show an dialog with details
        /// </summary>
        /// <param name="ex"></param>
        private void HandleError(Exception ex)
        {
            //The MLogics and MLgocis.Forms namespaces have a lot of very useful facilities, 
            //that you can use, such as Logging, exception handling, Language Switching etc. 
            //in this case we use the default Exception dialog to show an helpful message
            var form = new MLogics.Forms.ExceptionDialog(ex);
            form.ShowDialog();
        }

        /// <summary>
        /// Values gotten from the communication module, can be NULL, for example 
        /// if the driver is not connected yet. 
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        private string FormatValueToString(object value)
        {
            if (value == null) return "<null>"; //if it really is NULL, we write that to the UI
            return value.ToString(); //Otherwise we let the Value format 
        }

        private void buttonConnect_Click(object sender, EventArgs e)
        {
            try
            {
                //If we connect to the Communication Module, We need to have some Commandline parameters 
                //the Communication Module passes to us, when it start this module
                //Here we fake the commandline with our own predefined string 
                //AppIntf.ConnectToGp(Environment.CommandLine);
                ModCon.Connect();

                //From an Module Connection we can construct and Client, Server or an Config provider, all running on the same Moudle Connection.
                //in this example we only want an Variable client.
                //but the Module Connection MUST be connected first
                AppIntf = new Client(ModCon);
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
            finally
            {
                RefreshConnectionStateView();
            }
        }

        private void buttonDisconnect_Click(object sender, EventArgs e)
        {
            try
            {
                buttonUnsubscribe_Click(this, null);

                //Before disconnecting, we have to Destruct our client
                if (AppIntf != null)
                {
                    AppIntf.Dispose(); //if we call Dispose, also all outstanding Groups and Variables get Disposed automatically
                    AppIntf = null;  //this also makes our "SubscribedValueItem" invalid.
                }
        
                ModCon.Disconnect();
                SubscribedVariable = null; //this got Disposed automatically by the "Disconnect" call, so we also have to discard it
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
            finally
            {
                RefreshConnectionStateView();
            }
        }

        private void buttonWrite_Click(object sender, EventArgs e)
        {
            try
            {
                //To write values, we can use one of the convenient functions of the Application interface
                //we could also write to SubscribedValueItem.Value instead. 
                //Conversion of the Value into the appropriate type is handled by the library automatically
                AppIntf.SetVariableValue(textBoxReadWriteVarName.Text, textBoxReadWriteValue.Text);
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
        }

        private void buttonRead_Click(object sender, EventArgs e)
        {
            try
            {
                //We can use one of the "GetValueItemValue" methods, which are an easy and convenient way 
                //to get values from Variables
                //When we get an value, it might be null, so we have to prepare for that  
                textBoxReadWriteValue.Text = FormatValueToString(AppIntf.GetVariableValue(textBoxReadWriteVarName.Text));
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
        }

        private void buttonSubscribe_Click(object sender, EventArgs e)
        {
            try
            {
                //First we create an Client Variable. We have to Cache an reference to it, so that it does not
                //get garbage collected while we are still subscribed. 
                SubscribedVariable =  AppIntf.CreateVariable(textBoxSubscribeVarName.Text);

                //We subscribe to the Event
                SubscribedVariable.ValueChanged +=SubscribedValueItem_OnValue;

                //And set the Values mode to "Advised" so that it sends Value Change notifications to us
                SubscribedVariable.Advise(true);
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
            finally
            {
                RefreshSubscribedView();
            }
        }

        private void SubscribedValueItem_OnValue(object sender, VariableValueOperationEventArgs e)
        {
            //The Notification might get invoked by an different thread than the Main UI thread, 
            //so we have to Invoke back to the UI thread
            this.BeginInvoke(new Action (() => 
            {
                textBoxSubscribeValue.Text = FormatValueToString(e.Value); 
            }));
        }

        private void buttonUnsubscribe_Click(object sender, EventArgs e)
        {
            try
            {
                //First we deactivate the Advised mode
                SubscribedVariable.Deactivate();

                //Then we can un-subscribe from the event, since it will never be fired again
                SubscribedVariable.ValueChanged -=SubscribedValueItem_OnValue;

                //We have to Tell the Application interface, that we do not wish to use this ValueItem anymore
                SubscribedVariable.Group.Dispose(); //This also disposes the variable itself

                //since we disposed it, we discard its handle
                SubscribedVariable = null;
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
            finally
            {
                RefreshSubscribedView();
            }
        }

        private void GetSetVarsDialog_FormClosed(object sender, FormClosedEventArgs e)
        {
            buttonDisconnect_Click(this, null);
        }
    }
}
