﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using MLogics.VisXpert.Online.Client;//This is the Namespace where most of your Variable values functionality lives

namespace Example_Get_or_Set_Variables
{
    public partial class GetSetVarsDialog : Form
    {
        /// <summary>
        /// this is the Application Interface V1. It represents an connection to the Communication Module and
        /// allows you to exchange data with the Communication module. 
        /// It has Facilities to Manipulate Variable Values and to Subscribe to Events and Value changes
        /// 
        /// Usually you will have one instance of this Per Application. 
        /// To facilitate this, you can also use the ApplicationIntf.Instance, which is an Singleton
        /// </summary>
        private ApplicationIntf AppIntf = new ApplicationIntf();

        public GetSetVarsDialog()
        {
            InitializeComponent();
            RefreshConnectionStateView();
            RefreshSubscribedView();
        }

        private void RefreshConnectionStateView()
        {
            groupBoxReadWrite.Enabled = AppIntf.Connected;
            groupBoxSubscribe.Enabled = AppIntf.Connected;
            buttonDisconnect.Enabled = AppIntf.Connected;
            buttonConnect.Enabled = !AppIntf.Connected;

            if (AppIntf.Connected) labelConnectionState.Text = "Connected";
            else labelConnectionState.Text = "Disconnected";
        }

        /// <summary>
        /// We need to hold on tho an Handle of the ValueItem we subscribed to. 
        /// This way we know if we have an subscribed item or not, but we also keep the item 
        /// Alive, so it does not get Garbage collected
        /// </summary>
        ValueItem SubscribedValueItem;

        private void RefreshSubscribedView()
        {
            buttonSubscribe.Enabled = SubscribedValueItem == null;
            buttonUnsubscribe.Enabled = SubscribedValueItem != null;
        }

        /// <summary>
        /// This represents our generic error handler. Of course you should differentiate exception by type and
        /// Recover where possible, but in our demonstration we only show an dialog with details
        /// </summary>
        /// <param name="ex"></param>
        private void HandleError(Exception ex)
        {
            //The MLogics and MLgocis.Forms namespaces have a lot of very useful facilities, 
            //that you can use, such as Logging, exception handling, Language Switching etc. 
            //in this case we use the default Exception dialog to show an helpful message
            var form = new MLogics.Forms.ExceptionDialog(ex);
            form.ShowDialog();
        }

        /// <summary>
        /// Values gotten from the communication module, can be NULL, for example 
        /// if the driver is not connected yet. 
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        private string FormatValueToString(object value)
        {
            if (value == null) return "<null>";
            return value.ToString();
        }

        private void buttonConnect_Click(object sender, EventArgs e)
        {
            try
            {
                //If we connect to the Communication Module, We need to have some Commandline parameters 
                //the Communication Module passes to us, when it start this module
                //Here we fake the commandline with our own predefined string 
                //AppIntf.ConnectToGp(Environment.CommandLine);
                AppIntf.ConnectToGp("/N=\"MeasurementView 1\" /T=\"MeasurementView\" /A /P=\"\"");
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
            finally
            {
                RefreshConnectionStateView();
            }
        }

        private void buttonDisconnect_Click(object sender, EventArgs e)
        {
            try
            {
                //if we call Disconnect, also all outstanding Groups and Variables get Disposed automatically
                //this also makes our "SubscribedValueItem" invalid.
                AppIntf.DisconnectFromGp();

                SubscribedValueItem = null; //this got Disposed automatically by the "Disconnect" call, so we also have to discard it
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
            finally
            {
                RefreshConnectionStateView();
            }
        }

        private void buttonWrite_Click(object sender, EventArgs e)
        {
            try
            {
                //To write values, we can use one of the convenient functions of the Application interface
                //we could also write to SubscribedValueItem.Value instead. 
                //Conversion of the Value into the appropriate type is handled by the library automatically
                AppIntf.SetValueItemValue(textBoxReadWriteVarName.Text, textBoxReadWriteValue.Text);
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
        }

        private void buttonRead_Click(object sender, EventArgs e)
        {
            try
            {
                //We can use one of the "GetValueItemValue" methods, which are an easy and convenient way 
                //to get values from Variables
                //When we get an value, it might be null, so we have to prepare for that  
                textBoxReadWriteValue.Text = FormatValueToString(AppIntf.GetValueItemValue(textBoxReadWriteVarName.Text));
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
        }

        private void buttonSubscribe_Click(object sender, EventArgs e)
        {
            try
            {
                //First we create an Value item. We have to Cache an reference to it, so that it does not
                //get garbage collected while we are still subscribed. 
                SubscribedValueItem =  AppIntf.CreateValueItem(textBoxSubscribeVarName.Text);

                //We subscribe to the Event
                SubscribedValueItem.OnValue +=SubscribedValueItem_OnValue;

                //And set the Values mode to "Advised" so that it sends Value Change notifications to us
                SubscribedValueItem.Mode = MLogics.VisXpert.Online.Client.Types.ItemMode.Advised;
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
            finally
            {
                RefreshSubscribedView();
            }
        }

        private void SubscribedValueItem_OnValue(object sender, EventArgs e)
        {
            //The Notification might get invoked by an different thread than the Main UI thread, 
            //so we have to Invoke back to the UI thread
            this.BeginInvoke(new Action (() => 
            {
                textBoxSubscribeValue.Text = FormatValueToString(SubscribedValueItem.Value); 
            }));
        }

        private void buttonUnsubscribe_Click(object sender, EventArgs e)
        {
            try
            {
                //First we deactivate the Advised mode
                SubscribedValueItem.Mode = MLogics.VisXpert.Online.Client.Types.ItemMode.Inactive;

                //Then we can un-subscribe from the event, since it will never be fired again
                SubscribedValueItem.OnValue -=SubscribedValueItem_OnValue;

                //We have to Tell the Application interface, that we do not wish to use this ValueItem anymore
                AppIntf.RemoveValueItem(SubscribedValueItem);

                //since we disposed it, we discard its handle
                SubscribedValueItem = null;
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
            finally
            {
                RefreshSubscribedView();
            }
        }

        private void GetSetVarsDialog_FormClosed(object sender, FormClosedEventArgs e)
        {
            //We can omit an Try/Catch block, because Disconnect will never Throw an exception
            AppIntf.DisconnectFromGp();
        }
    }
}
