﻿using System;
using System.Collections.Generic;

namespace Example_Variable_Driver
{
    internal class AdvisedVariableManager
    {
        private const int MasModbusDataLength = 200;
        ModbusServerGrp grp;
        IList<ModbusServerVar> advisedVars = new List<ModbusServerVar>();
        IEnumerable<ModbusRequest> reqs;
        object varLock = new object();

        public static MLogics.Logging.Logger Log { get; } = new MLogics.Logging.Logger(nameof(AdvisedVariableManager));

        public AdvisedVariableManager(ModbusServerGrp grp)
        {
            this.grp = grp;
        }
        public TimeSpan AverageVarResponseTime { get; private set; }

        #region Update Vars
        public void StartPeriodicUpdates()
        {
            if (periodicthread != null) return; //it is already running
            wait.Reset();
            periodicthread = new System.Threading.Thread(DoPeriodicUpdate);
            periodicthread.Start();
            Log.Debug("Periodic update started for Group {0}", grp.GroupName);
        }

        public void StopPeriodicUpdates()
        {
            wait.Set();
        }

        private System.Threading.Thread periodicthread;
        private System.Threading.ManualResetEvent wait = new System.Threading.ManualResetEvent(false);
        private TimeSpan TickDuration = TimeSpan.FromMilliseconds(100);
        private void DoPeriodicUpdate()
        {
            do
            {
                if (reqs == null) continue;
                lock (varLock)
                {
                    if (grp.MbCon == null) break; //We somehow missed the StopPeriodicUpdates call
                    foreach (var req in reqs)
                    {
                        try
                        {
                            if (DateTime.Now.Subtract(req.LastUpdate) > req.RefreshInterval)
                            {
                                var sw = new System.Diagnostics.Stopwatch();
                                sw.Start();
                                req.ReadFromMaster(grp.MbCon, grp.ModbusID);
                                AverageVarResponseTime = TimeSpan.FromMilliseconds(sw.Elapsed.TotalMilliseconds / req.Vars.Count);
                            }
                        }
                        catch (Exception ex)
                        {
                            Log.Error("Error while updating Modbus request {0} with error {1}", req, ex);
                        }
                    }
                }

            } while (!wait.WaitOne((int)TickDuration.TotalMilliseconds));
            periodicthread = null;
            Log.Debug("Periodic update stopped for Group {0}", grp.GroupName);
        }
        #endregion

        #region Manage List and Optimize Requests
        public void AddVariable(ModbusServerVar variable)
        {
            lock(varLock)
            {
                if (advisedVars.Contains(variable)) return;
                advisedVars.Add(variable);
                reqs = BuildModbusRequests(advisedVars);
            }
        }

        public void RemoveVariable(ModbusServerVar variable)
        {
            lock (varLock)
            {
                advisedVars.Remove(variable);
                reqs = BuildModbusRequests(advisedVars);
            }
        }

        private IEnumerable<ModbusRequest> BuildModbusRequests(IEnumerable<ModbusServerVar> vars)
        {
            var reqs = new List<ModbusRequest>();

            foreach (var var in vars)
            {
                var varReq = new ModbusRequest(var);
                var found = false;

                foreach (var req in reqs)
                {
                    if (req.ModbusDataArea != varReq.ModbusDataArea) continue;
                    if (req.RefreshInterval.TotalMilliseconds != varReq.RefreshInterval.TotalMilliseconds) continue;

                    //if it is fully inside an existing Request, we can just use that one
                    if (varReq.IsFullyInsideOf(req))
                    {
                        req.Vars.Add(var);
                        found = true;
                        break;
                    }

                    //Now check if it cou8ld be inside another one
                    var canidate = new ModbusRequest(req);
                    canidate.AddAndAdaptRequest(var);
                    if (canidate.RegisterCount <= MasModbusDataLength)
                    {
                        req.AddAndAdaptRequest(var);
                        found = true;
                        break;
                    }
                }

                //We could not match it, so we create an new Request
                if (!found)
                {
                    reqs.Add(new ModbusRequest(var));
                }
            }

            return reqs;
        }
        #endregion
    }
}
