﻿using System;

namespace Example_Variable_Driver
{
    /// <summary>
    /// Represents an Modbus Connection to an Modbus device. it implements an Reconnection strategy and monitors the connection
    /// it also implements "Chunking", meaning it breaks up big messages into smaller ones, as to not exceed the modbus limits
    /// </summary>
    internal class ModbusConnection
    {
        private NModbus.IModbusMaster ModbusMaster { get; set; }
        private System.Net.Sockets.TcpClient TcpConnection { get; set; }
        private static MLogics.Logging.Logger Log { get; } = new MLogics.Logging.Logger(nameof(ModbusConnection));

        public string Address { get; private set; }
        public int Port { get; private set; }
        public byte ModbusID { get; private set; }
        public TimeSpan Timeout { get; private set; } = TimeSpan.FromSeconds(10);
        public TimeSpan RetryTime { get; private set; } = TimeSpan.FromSeconds(30);
        private const ushort MaximumDiscreteRequestResponseSize = 2040;
        private const ushort MaximumRegisterRequestResponseSize = 127;

        /// <summary>
        /// Use that to Synchronize usage to
        /// </summary>
        public object SyncRoot { get; } = new object();

        public ModbusConnection(string Address, int Port, TimeSpan Timeout, TimeSpan RetryTime)
        {
            this.Address =  Address;
            this.Port = Port;
            this.Timeout = Timeout;
            this.RetryTime = RetryTime;
        }

        public event EventHandler ConnectionFailed;
        public event EventHandler ConnectionReestablished;

        public void Connect()
        {
            if (ModbusMaster != null) return;

            var Factory = new NModbus.ModbusFactory();
            TcpConnection = new System.Net.Sockets.TcpClient();
            TcpConnection.ReceiveTimeout = (int)Timeout.TotalMilliseconds;
            TcpConnection.SendTimeout = (int)Timeout.TotalMilliseconds;

            try
            {
                TcpConnection.Connect(Address, Port);
                ModbusMaster = Factory.CreateMaster(TcpConnection);
                Log.Debug("Connected to IP Address {0}", Address);
            }
            catch (Exception ex)
            {
                Log.Error("Failed to Connect to IP Address {0} with error {1}", Address, ex);
                BeginReconnect();
                throw; //we have to let the Exception to bubble up so the correct message can be sent to the Comm module   
            }
        }

        public void Disconnect()
        {
            StopReconnection();
            if (ModbusMaster is object)
            {
                ModbusMaster.Dispose();
                ModbusMaster = null;
            }
            if (TcpConnection is object)
            {
                TcpConnection.Close();
                TcpConnection.Dispose();
                TcpConnection = null;
            }
        }

        public bool IsConnected
        {
            get 
            {
                if (TcpConnection is null) return false;
                return TcpConnection.Connected;
            }
        }

        #region Modbus helpers
        /// <summary>
        /// Log an exception to the Trace listener and also log the Function name where it happened 
        /// </summary>
        /// <param name="ex"></param>
        /// <param name="funcName"></param>
        private void LogException(Exception ex, [System.Runtime.CompilerServices.CallerMemberName]string funcName = "")
        {
            Log.Error("Error during function '{0}' with error {1}",funcName, ex);
        }

        /// <summary>
        /// Runs an Modbus request, and depending on the type of Exception either sets the request on error by throwing the exception, 
        /// or triggers an Reconnect
        /// </summary>
        /// <param name="req"></param>
        /// <param name="funcName"></param>
        private void RunGuardedModbusRequest(Action req, [System.Runtime.CompilerServices.CallerMemberName] string funcName = "")
        {
            try
            {
                lock (SyncRoot)
                {
                    CheckModbusConnection();
                    req.Invoke();
                }
            }
            catch (Exception ex)
            {
                HandleModbusException(ex, funcName);
            }
        }

        /// <summary>
        /// Runs an Modbus request, and depending on the type of Exception either sets the request on error by throwing the exception, 
        /// or triggers an Reconnect
        /// </summary>
        /// <param name="req"></param>
        /// <param name="funcName"></param>
        private bool[] RunGuardedModbusRequest(Func<bool[]> req, [System.Runtime.CompilerServices.CallerMemberName] string funcName = "")
        {
            try
            {
                lock (SyncRoot)
                {
                    CheckModbusConnection();
                    return req.Invoke();
                }
            }
            catch (Exception ex)
            {
                HandleModbusException(ex, funcName);
                return null;
            }
        }

        /// <summary>
        /// Runs an Modbus request, and depending on the type of Exception either sets the request on error by throwing the exception, 
        /// or triggers an Reconnect
        /// </summary>
        /// <param name="req"></param>
        /// <param name="funcName"></param>
        private ushort[] RunGuardedModbusRequest(Func<ushort[]> req, [System.Runtime.CompilerServices.CallerMemberName] string funcName = "")
        {
            try
            {
                lock (SyncRoot)
                {
                    CheckModbusConnection();
                    return req.Invoke();
                }
            }
            catch (Exception ex)
            {
                HandleModbusException(ex, funcName);
                return null;
            }
        }

        /// <summary>
        /// Before running an modbus request, it checks if an Connection is established in the first place
        /// </summary>
        /// <exception cref="ArgumentNullException"></exception>
        /// <exception cref="ArgumentException"></exception>
        private void CheckModbusConnection()
        {
            if (ReconnectTimer is object) throw new System.IO.IOException("Reconnection is currently running, but no connection established yet");
            if (ModbusMaster is null) throw new System.IO.IOException("No connection is established yet");
            if (TcpConnection is null) throw new System.IO.IOException("No connection is established yet");
            if (!TcpConnection.Connected) throw new System.IO.IOException("No connection is established yet");
        }

        private void HandleModbusException(Exception ex, [System.Runtime.CompilerServices.CallerMemberName] string funcName = "")
        {
            if (ex is TimeoutException)
            {
                LogException(ex, funcName);
                BeginReconnect();
                throw ex;
            }
            else if (ex is NModbus.SlaveException)
            {
                LogException(ex, funcName);
                throw ex;
            }
            else if(ex is NModbus.InvalidModbusRequestException)
            {
                LogException(ex, funcName);
                throw ex;
            }
            else if(ex is Exception)
            {
                LogException(ex, funcName);
                BeginReconnect();
                throw ex;
            }
        }
        #endregion

        #region Modbus

        public bool[] ReadCoils(byte slaveAddress, ushort startAddress, ushort numberOfPoints)
        {
            return RunGuardedModbusRequest(() =>
            {
                if (numberOfPoints <= MaximumDiscreteRequestResponseSize)
                {
                    return ModbusMaster.ReadCoils(slaveAddress, startAddress, numberOfPoints);
                }
                else 
                {
                    bool[] resp = new bool[numberOfPoints];
                    for(ushort i = 0; i < numberOfPoints; i+=MaximumDiscreteRequestResponseSize)
                    {
                        var partResp = ModbusMaster.ReadCoils(  slaveAddress, 
                                                                (ushort)(startAddress + i),
                                                                (ushort)Math.Min(numberOfPoints - i, MaximumDiscreteRequestResponseSize));
                        Buffer.BlockCopy(partResp,0, resp, i, partResp.Length);
                    }

                    return resp;
                }
            });
        }

        public bool[] ReadInputs(byte slaveAddress, ushort startAddress, ushort numberOfPoints)
        {
            return RunGuardedModbusRequest(() =>
            {
                if (numberOfPoints <= MaximumDiscreteRequestResponseSize)
                {
                    return ModbusMaster.ReadInputs(slaveAddress, startAddress, numberOfPoints);
                }
                else
                {
                    bool[] resp = new bool[numberOfPoints];
                    for (ushort i = 0; i < numberOfPoints; i+=MaximumDiscreteRequestResponseSize)
                    {
                        var partResp = ModbusMaster.ReadInputs(slaveAddress,
                                                                (ushort)(startAddress + i),
                                                                (ushort)Math.Min(numberOfPoints - i, MaximumDiscreteRequestResponseSize));
                        Buffer.BlockCopy(partResp, 0, resp, i, partResp.Length);
                    }

                    return resp;
                }
            });
        }

        public ushort[] ReadHoldingRegisters(byte slaveAddress, ushort startAddress, ushort numberOfPoints)
        {
            return RunGuardedModbusRequest(() =>
            {
                if (numberOfPoints <= MaximumRegisterRequestResponseSize)
                {
                    return ModbusMaster.ReadHoldingRegisters(slaveAddress, startAddress, numberOfPoints);
                }
                else
                {
                    ushort[] resp = new ushort[numberOfPoints];
                    for (ushort i = 0; i < numberOfPoints; i+=MaximumDiscreteRequestResponseSize)
                    {
                        var partResp = ModbusMaster.ReadHoldingRegisters(slaveAddress,
                                                                (ushort)(startAddress + i),
                                                                (ushort)Math.Min(numberOfPoints - i, MaximumDiscreteRequestResponseSize));
                        Buffer.BlockCopy(partResp, 0, resp, i, partResp.Length);
                    }

                    return resp;
                }
            });
        }

        public ushort[] ReadInputRegisters(byte slaveAddress, ushort startAddress, ushort numberOfPoints)
        {
            return RunGuardedModbusRequest(() =>
            {
                if (numberOfPoints <= MaximumRegisterRequestResponseSize)
                {
                    return ModbusMaster.ReadInputRegisters(slaveAddress, startAddress, numberOfPoints);
                }
                else
                {
                    ushort[] resp = new ushort[numberOfPoints];
                    for (ushort i = 0; i < numberOfPoints; i+=MaximumDiscreteRequestResponseSize)
                    {
                        var partResp = ModbusMaster.ReadInputRegisters(slaveAddress,
                                                                (ushort)(startAddress + i),
                                                                (ushort)Math.Min(numberOfPoints - i, MaximumDiscreteRequestResponseSize));
                        Buffer.BlockCopy(partResp, 0, resp, i, partResp.Length);
                    }

                    return resp;
                }
            });
        }

        public void WriteSingleCoil(byte slaveAddress, ushort coilAddress, bool value)
        {
            RunGuardedModbusRequest(() =>
            {
                ModbusMaster.WriteSingleCoil(slaveAddress, coilAddress, value);
            });
        }

        public void WriteMultipleRegisters(byte slaveAddress, ushort startAddress, ushort[] data)
        {
            RunGuardedModbusRequest(() =>
            {
                if (data.Length <= MaximumDiscreteRequestResponseSize)
                {
                    ModbusMaster.WriteMultipleRegisters(slaveAddress, startAddress, data);
                }
                else
                {
                    for (ushort i = 0; i < data.Length; i+=MaximumDiscreteRequestResponseSize)
                    {
                        var parData = new ushort[Math.Min(data.Length - i, MaximumDiscreteRequestResponseSize)];
                        Buffer.BlockCopy(data, i, parData, 0, parData.Length);
                        ModbusMaster.WriteMultipleRegisters(slaveAddress, (ushort)(startAddress + i), parData);
                    }
                }
            });
        }

        #endregion

        #region Reconnection
        private System.Timers.Timer ReconnectTimer;
        private void BeginReconnect()
        {
            if (ReconnectTimer is object) return; //Reconnection already running
            ReconnectTimer = new System.Timers.Timer(RetryTime.TotalMilliseconds);
            ReconnectTimer.Elapsed += ReconnectTimer_Elapsed;
            TcpConnection.Close();
            ReconnectTimer.Start();
            ConnectionFailed?.Invoke(this, EventArgs.Empty);
        }

        private void TryReconnect()
        {
            try
            {
                lock (SyncRoot)
                {
                    if (TcpConnection is null)
                    {
                        StopReconnection();
                        return;
                    }
                    if (TcpConnection.Client != null && TcpConnection.Connected)
                    {
                        StopReconnection();
                        return;
                    }
                    TcpConnection.Close();
                    TcpConnection = null;
                    if (ModbusMaster!= null) ModbusMaster.Dispose();
                    ModbusMaster = null;
                    Connect();
                    StopReconnection();
                    ConnectionReestablished?.Invoke(this, EventArgs.Empty);
                }
            }
            catch (Exception ex)
            {
                Log.Debug("Failed to reconnect to {0}:{1} with error {2}", Address.ToString(),Port, ex);
            }
        }

        private void ReconnectTimer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            TryReconnect();
        }

        private void StopReconnection()
        {
            if (ReconnectTimer == null) return;
            lock (SyncRoot)
            {
                ReconnectTimer.Stop();
                ReconnectTimer.Elapsed -= ReconnectTimer_Elapsed;
                ReconnectTimer = null;
            }
        }
        #endregion
    }
}
