﻿using System;
using System.Collections.Generic;

namespace Example_Variable_Driver
{
    /// <summary>
    /// Represents an single request that gets sent to the wire. it contains one or several variables that
    /// will get their data from this request
    /// </summary>
    internal class ModbusRequest
    {
        public static MLogics.Logging.Logger Log { get; } = new MLogics.Logging.Logger(nameof(ModbusRequest));

        public ModbusDataAreas ModbusDataArea { get; set; } = ModbusDataAreas.HoldingRegisters;
        public ushort RegisterStart { get; set; } = ushort.MaxValue;
        public ushort RegisterCount { get; set; } = 0;
        public ushort RegisterEnd
        {
            get { return (ushort)(RegisterStart + RegisterCount); }
            set { RegisterCount = (ushort)(value - RegisterStart); }
        }
        public TimeSpan RefreshInterval { get; set; }
        public DateTime LastUpdate { get; set; } = DateTime.Now;

        public IList<ModbusServerVar> Vars { get; set; } = new List<ModbusServerVar>();

        /// <summary>
        /// Does it intersect with the Register range of another request
        /// </summary>
        /// <param name="mbRequest"></param>
        /// <returns></returns>
        public bool Intersects(ModbusRequest mbRequest)
        {
            if (this.RegisterStart <= mbRequest.RegisterEnd)
            {
                return (this.RegisterEnd >= mbRequest.RegisterStart);
            }
            else if (this.RegisterEnd >= mbRequest.RegisterStart)
            {
                return (this.RegisterStart <= mbRequest.RegisterEnd);
            }

            return false;
        }

        /// <summary>
        /// is the given Requests register range fully inside of this one
        /// </summary>
        /// <param name="mbRequest"></param>
        /// <returns></returns>
        public bool IsFullyInside(ModbusRequest mbRequest)
        {
            if (this.RegisterStart >= mbRequest.RegisterStart)
            {
                return this.RegisterEnd <= mbRequest.RegisterEnd;
            }

            return false;
        }

        /// <summary>
        /// is register range this one fully inside of the given requests
        /// </summary>
        /// <param name="mbRequest"></param>
        /// <returns></returns>
        public bool IsFullyInsideOf(ModbusRequest mbRequest)
        {
            return mbRequest.IsFullyInside(this);
        }

        /// <summary>
        /// Add an new Variable and Recalculate the Range
        /// </summary>
        /// <param name="var"></param>
        public void AddAndAdaptRequest(ModbusServerVar var)
        {
            var beforeEnd = RegisterEnd;

            if (RegisterStart > var.RegisterStart || RegisterStart ==  ushort.MaxValue)
            {
                RegisterStart = var.RegisterStart;
                RegisterEnd = beforeEnd;
            }
            if (RegisterEnd < (var.RegisterStart + var.RegisterCount))
            {
                RegisterEnd = beforeEnd;
            }
        }

        public void ReadFromMaster(ModbusConnection master, byte ModbusID)
        {
            if (master == null) return;
            LastUpdate = DateTime.Now;
            try
            {
                switch (ModbusDataArea)
                {
                    case ModbusDataAreas.DiscreteInputs:
                        var bits1 = master.ReadInputs(ModbusID, RegisterStart, RegisterCount);
                        ReadFromMasterAsDiscretes(bits1);
                        break;

                    case ModbusDataAreas.Coils:
                        var bits2 = master.ReadCoils(ModbusID, RegisterStart, RegisterCount);
                        ReadFromMasterAsDiscretes(bits2);
                        break;

                    case ModbusDataAreas.HoldingRegisters:
                        var regs1 = master.ReadHoldingRegisters(ModbusID,RegisterStart,RegisterCount);
                        ReadFromMasterAsRegisters(regs1);
                        break;

                    case ModbusDataAreas.InputRegisters:
                        var regs2 = master.ReadInputRegisters( ModbusID,RegisterStart,RegisterCount);
                        ReadFromMasterAsRegisters(regs2);
                        break;

                    default:
                        return;
                }            
            }
            catch (Exception ex)
            {
                Log.Error("Error while reading request {0} with error {1}", this.ToString(), ex);
                foreach (var Var in Vars) Var.SetError(ex);
            }
        }

        public void ReadFromMasterAsRegisters(ushort[] regs)
        {
            if (regs.Length != RegisterCount)
            {
                throw new ArgumentOutOfRangeException(String.Format("requested {0} registers, but received {1} registers!", RegisterCount, regs.Length));
            }

            foreach (var Var in Vars)
            {
                try
                {
                    ushort[] VarRegs = new ushort[Var.RegisterCount];
                    Buffer.BlockCopy(regs, Var.RegisterStart - RegisterStart, VarRegs, 0, Var.RegisterCount); //it is the offset inside the Requests offset
                    Var.ValueFromRegisters(VarRegs);
                    Var.ResetError();
                }
                catch (Exception varex)
                {
                    Log.Error("Error while processing variable  {0} with error {1}", Var.ToString(), varex);
                    Var.SetError();
                }
            }
        }

        public void ReadFromMasterAsDiscretes(bool[] regs)
        {
            if (regs.Length != RegisterCount)
            {
                throw new ArgumentOutOfRangeException(String.Format("requested {0} registers, but received {1} registers!", RegisterCount, regs.Length));
            }

            foreach (var Var in Vars)
            {
                try
                {
                    Var.ValueFromDiscrete(regs[Var.RegisterStart - RegisterStart]); //it is the offset inside the Requests offset
                    Var.ResetError();
                 }
                catch (Exception varex)
                {
                    Log.Error("Error while processing variable  {0} with error {1}", Var.ToString(), varex);
                    Var.SetError();
                }
            }
        }

        public ModbusRequest()
        { }

        public ModbusRequest(ModbusServerVar var)
        {
            RegisterStart = var.RegisterStart;
            RegisterCount = var.RegisterCount;
            ModbusDataArea = var.ModbusDataArea;
            RefreshInterval = var.RefreshInterval;
            Vars.Add(var);
        }

        public ModbusRequest(ModbusRequest req)
        {
            RegisterStart = req.RegisterStart;
            RegisterCount = req.RegisterCount;
            ModbusDataArea = req.ModbusDataArea;
            RefreshInterval = req.RefreshInterval;
        }

        public override string ToString()
        {
            return String.Format("{0} from {1} to {2}", ModbusDataArea.ToString(), RegisterStart, RegisterEnd);
        }
    }
}
