﻿using MLogics.VisXpert.Online.Server;
using System;

namespace Example_Variable_Driver
{
    /// <summary>
    /// An Server group holds one single connection to one single Modbus Device
    /// </summary>
    internal class ModbusServerGrp: ServerGroup
    {
        private static MLogics.Logging.Logger Log { get; } = new MLogics.Logging.Logger(nameof(ModbusServerGrp));

        public ModbusServerGrp(string GroupName):base(GroupName)
        {
            AdvisedVarMgr = new AdvisedVariableManager(this);
        }

        public ModbusServerGrp(Configuration.ModbusGroupConfig GroupCfg) : base(GroupCfg.GroupName)
        {
            AdvisedVarMgr = new AdvisedVariableManager(this);
            Address = GroupCfg.Address;
            Port = GroupCfg.Port;
            ModbusID = GroupCfg.ModbusID;
            Timeout = GroupCfg.Timeout; 
            RetryTime = GroupCfg.RetryTime;
        }

        protected override ServerVariable CreateNewServerVar(string name)
        {
            return new ModbusServerVar(name);
        }

        public string Address { get; set; }
        public int Port { get; set; }   
        public byte ModbusID { get; set; }
        public TimeSpan Timeout { get; set; } = TimeSpan.FromSeconds(10);
        public TimeSpan RetryTime { get; set; } = TimeSpan.FromSeconds(30);
        public TimeSpan AverageVarResponseTime { get { return AdvisedVarMgr.AverageVarResponseTime; } }

        internal ModbusConnection MbCon;
        private AdvisedVariableManager AdvisedVarMgr;

        private void SetVarsError()
        {
            foreach (ModbusServerVar Var in Variables) Var.SetError();
        }
        private void ResetVarError()
        {
            foreach (ModbusServerVar Var in Variables) Var.ResetError();
        }

        protected override bool OnConnect()
        {
            try
            {
                if (MbCon is null)
                {
                    MbCon = new ModbusConnection(Address, Port, Timeout, RetryTime);
                    MbCon.ConnectionFailed += (a,b) => { SetVarsError(); AdvisedVarMgr.StopPeriodicUpdates(); };
                    MbCon.ConnectionReestablished += (a,b) => { ResetVarError(); AdvisedVarMgr.StartPeriodicUpdates(); };
                }
                MbCon.Connect();
            }
            catch (Exception ex)
            {
                Log.Error("Failed to Connect to IP Address {0} with error {1}", Address, ex);
                throw; //we have to let the Exception to bubble up so the correct message can be sent to the Comm module   
            }

            return base.OnConnect();
        }
        protected override bool OnDisconnect()
        {
            AdvisedVarMgr.StopPeriodicUpdates();
            MbCon.Disconnect();
            return base.OnDisconnect();
        }

        protected override void OnFirstVariableConnected(ServerVariable variable)
        {
            Log.Debug("first Variable is connecting to Group {0}", GroupName);
            AdvisedVarMgr.StartPeriodicUpdates();
        }
        protected override void OnLastVariableDisconnected(ServerVariable variable)
        {
            Log.Debug("last Variable disconnected from Group {0}", GroupName);
            AdvisedVarMgr.StopPeriodicUpdates();
        }

        /// <summary>
        /// Warning! this will be called twice. Once if Advise(All=False) and then again with Advise(All=True)
        /// </summary>
        /// <param name="variable"></param>
        protected override void OnVariableAdvised(ServerVariable variable)
        {
            AdvisedVarMgr.AddVariable((ModbusServerVar)variable);
            base.OnVariableAdvised(variable);
        }
        protected override void OnVariableUnadvised(ServerVariable variable)
        {
            AdvisedVarMgr.RemoveVariable((ModbusServerVar)variable);
            base.OnVariableUnadvised(variable);
        }
    }
}