﻿using MLogics.VisXpert.Online.Server;
using System;
using System.Text;

namespace Example_Variable_Driver
{
    internal partial class ModbusServerVar : ServerVariable
    {
        public static MLogics.Logging.Logger Log { get; } = new MLogics.Logging.Logger(nameof(ModbusServerVar));

        public ModbusServerVar(string VarName) : base(VarName)
        { }

        public ModbusServerVar(Configuration.ModbusVarConfig VarConfig) : base(VarConfig.VariableName)
        {
            ModbusDataArea = VarConfig.ModbusDataArea;
            ModbusDataType = VarConfig.ModbusDataType;
            RegisterStart = (ushort)VarConfig.RegisterStart;
            RegisterCount = (ushort)VarConfig.RegisterCount;
            ReadOnly = VarConfig.ReadOnly;
            RefreshInterval = VarConfig.RefreshTime;
            ReadExpression = VarConfig.ReadExpression;
            WriteExpression = VarConfig.WriteExpression;
        }

        public ModbusDataAreas ModbusDataArea { get; set; } = ModbusDataAreas.HoldingRegisters;
        public ModbusDataTypes ModbusDataType { get; set; }
        public ushort RegisterStart { get; set; }
        public ushort RegisterCount { get; set; }
        public bool ReadOnly { get; set; }
        public TimeSpan RefreshInterval { get; set; }
        public string ReadExpression { get; set; }
        public string WriteExpression { get; set; }

        #region ServerVar Overrides
        protected override void OnWriteValueToDevice(object val)
        {
            var mbGrp = (ModbusServerGrp)Group;
            mbGrp.MbCon.WriteMultipleRegisters(mbGrp.ModbusID, RegisterStart, ValueToRegisters(val));
            base.OnWriteValueToDevice(val);
            Log.Debug("{0} Wrote value {1} to device", ToString(), val);
        }

        protected override void OnAdvised(bool val)
        {
            Log.Debug("{0} Advised", ToString());
            base.OnAdvised(val);
        }
        protected override void OnUnadvised()
        {
            Log.Debug("{0} Unadvised", ToString());
            base.OnUnadvised();
        }

        protected override object OnReadValueFromDevice()
        {
            var mbGrp = (ModbusServerGrp)Group;
            lock (mbGrp.MbCon)
            {
                try
                {
                    var Req = new ModbusRequest(this);
                    Req.ReadFromMaster(mbGrp.MbCon, mbGrp.ModbusID);
                    ResetError();
                    return CachedDeviceValue;
                }
                catch (Exception ex)
                {
                    Log.Error("Failed to read variable {0} with error {1}", this.ToString(), ex);
                    SetError();
                    throw;
                }
            }
        }
        #endregion

        internal void SetError()
        {
            IsError = true;
        }

        internal void SetError(Exception ex)
        {
            CurrentError  = ex;
        }

        internal void ResetError()
        {
            IsError = false;
        }

        internal void ValueFromRegisters(ushort[] registers)
        {
            if (ReadOnly)
                throw new NModbus.InvalidModbusRequestException("Value is Read-only", NModbus.SlaveExceptionCodes.IllegalFunction);

            //We get the Byte representation
            var bytes = RegistersToBytes(registers);

            //From there we extract the value
            var VxValue = BytesToValue(bytes);

            //Apply the Expression on it
            VxValue = ResolveExpression(VxValue, WriteExpression);

            //And sent the value to VisXpert
            CachedDeviceValue = VxValue;
        }

        internal void ValueFromDiscrete(bool boolval)
        {
            if (ReadOnly)
                throw new NModbus.InvalidModbusRequestException("Value is Read-only", NModbus.SlaveExceptionCodes.IllegalFunction);

            object VxValue = boolval;

            //Apply the Expression on it
            VxValue = ResolveExpression(VxValue, WriteExpression);

            //And sent value to VisXpert
            CachedDeviceValue = VxValue;
        }

        internal ushort[] ValueToRegisters(object ValueTobeWritten)
        {
            try
            {
                //We get the value from VisXpert
                var VxValue = ValueTobeWritten;

                //Apply the Expression on it
                VxValue = ResolveExpression(VxValue, ReadExpression);

                //Convert the Type to the Requested Modbus type and resolve it to bytes
                var bytes = ValueToBytes(VxValue);

                //Combine the Bytes into Modbus Word Registers
                return BytesToRegisters(bytes);
            }
            catch (Exception ex)
            {
                throw new NModbus.InvalidModbusRequestException(NModbus.SlaveExceptionCodes.SlaveDeviceFailure, ex);
            }
        }

        internal bool ValueToCoil(object ValueTobeWritten)
        {
            try
            {
                //We get the value from VisXpert
                var VxValue = ValueTobeWritten;

                //Apply the Expression on it
                VxValue = ResolveExpression(VxValue, ReadExpression);

                //Convert it to an Bool by using the default system functions
                //Value != 0 = TRUE
                return Convert.ToBoolean(VxValue);
            }
            catch (Exception ex)
            {
                throw new NModbus.InvalidModbusRequestException(NModbus.SlaveExceptionCodes.SlaveDeviceFailure, ex);
            }
        }

        private static ushort[] BytesToRegisters(byte[] bytes)
        {
            ushort[] registers = new ushort[(int)Math.Ceiling((double)bytes.Length/2)];
            Buffer.BlockCopy(bytes, 0, registers, 0, bytes.Length);
            return registers;
        }

        private byte[] ValueToBytes(object VxValue)
        {
            switch (ModbusDataType)
            {
                case ModbusDataTypes.int16:
                    return BitConverter.GetBytes(Convert.ToInt16(VxValue));
                case ModbusDataTypes.int32:
                    return BitConverter.GetBytes(Convert.ToInt32(VxValue));
                case ModbusDataTypes.real:
                    return BitConverter.GetBytes(Convert.ToSingle(VxValue));
                case ModbusDataTypes.text:

                    //we have to make sure that your string will fit into Registercount *2 (two chars per register)
                    var s = Convert.ToString(VxValue);
                    if (s.Length > RegisterCount *2) s = s.Substring(0, RegisterCount*2); //Limit to configured length

                    //We have to return an array of the requested size. If the string is smaller
                    //Then the rest are NULL bytes
                    byte[] bytes = new byte[RegisterCount*2];
                    for (int i = 0; i < bytes.Length; i++) bytes[i] = 0;

                    //We copy the ASCII bytes onto the buffer. The remaining bytes stay "0"
                    var ascii = ASCIIEncoding.ASCII.GetBytes(s);
                    Array.Copy(ascii, 0, bytes, 0, ascii.Length);

                    return bytes;

                case ModbusDataTypes.bit:
                    return BitConverter.GetBytes(Convert.ToInt16(VxValue) != 0);
                default:
                    return null;
            }
        }

        private static byte[] RegistersToBytes(ushort[] registers)
        {
            byte[] bytes = new byte[registers.Length  *2];
            Buffer.BlockCopy(registers, 0, bytes, 0, registers.Length);
            return bytes;
        }

        private object BytesToValue(byte[] bytes)
        {
            switch (ModbusDataType)
            {
                case ModbusDataTypes.int16:
                    return BitConverter.ToInt16(bytes, 0);
                case ModbusDataTypes.int32:
                    return BitConverter.ToInt32(bytes, 0);
                case ModbusDataTypes.real:
                    return BitConverter.ToSingle(bytes, 0);
                case ModbusDataTypes.text:
                    return ASCIIEncoding.ASCII.GetString(bytes);
                case ModbusDataTypes.bit:
                    return bytes[0] != 0;
                default:
                    return null;
            }
        }

        private object ResolveExpression(object VxValue, string Expression)
        {
            if (String.IsNullOrWhiteSpace(Expression)) return VxValue;

            var Eval = new MLogics.ExpressionEvaluation.ExpressionEvaluator();
            Eval.AddEnvironmentFunctions(new ExpressionFunctions());
            Eval.Parse(ReadExpression);
            Eval.Variables.V.Value = VxValue;

            return Eval.Value;
        }

        private class ExpressionFunctions
        {
            public bool GetBit(int value, int bitNo) { return MLogics.BitMagic.GetBit(value, bitNo); }
            public int SetBit(int value, int bitNo, bool bitvalue) { return MLogics.BitMagic.SetBit(value, bitNo, bitvalue); }

            public ushort SwapBytes(ushort x)
            {
                return (ushort)((ushort)((x & 0xff) << 8) | ((x >> 8) & 0xff));
            }

            public uint SwapBytes(uint x)
            {
                return ((x & 0x000000ff) << 24) +
                       ((x & 0x0000ff00) << 8) +
                       ((x & 0x00ff0000) >> 8) +
                       ((x & 0xff000000) >> 24);
            }

            public uint SwapWords(uint x)
            {
                return ((x & 0x0000ffff) << 16) +
                       ((x & 0xffff0000) >> 16);
            }

            public uint SwapRegisters(uint x)
            {
                return ((x & 0x0000ffff) << 16) +
                       ((x & 0xffff0000) >> 16);
            }

            public ulong SwapBytes(ulong value)
            {
                ulong uvalue = value;
                ulong swapped =
                     ((0x00000000000000FF) & (uvalue >> 56)
                     | (0x000000000000FF00) & (uvalue >> 40)
                     | (0x0000000000FF0000) & (uvalue >> 24)
                     | (0x00000000FF000000) & (uvalue >> 8)
                     | (0x000000FF00000000) & (uvalue << 8)
                     | (0x0000FF0000000000) & (uvalue << 24)
                     | (0x00FF000000000000) & (uvalue << 40)
                     | (0xFF00000000000000) & (uvalue << 56));
                return swapped;
            }
        }
    }
}
