﻿using BatchXpert_GP_Batch;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Resolve_VisXpert_Assemblies_from_Any_Directory
{
    internal static class Program
    {
        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main()
        {
            SetUpAssemblyResolve();
            AddVisXpertPathToSearchPath();

            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);
            Application.Run(new ResolveVisXpertAssemblies());
        }

        /// <summary>
        /// Some VisXpert Assemblies have Native Dependencies, which can not be resolved by the 
        /// .Net AssemblyResolve event, which only works for managed Assemblies. 
        /// In order to Resolve these, we could add the "VisXpert Bin" directory to the Current directory, since 
        /// the Current directory is part of the Windows Search Path. 
        /// 
        /// But there is a better version. Since Windows 8.1 you can manually add Directories to the Search Path, without
        /// touching the "PATH" environment Variable. 
        /// This "Additional Directory" is only active for this Process, and does not effect any other application. 
        /// </summary>
        private static void AddVisXpertPathToSearchPath()
        {
            //First we get the Current VisXpert bin directory path
            var VxPath = System.IO.Path.Combine(MLogics.VisXpert.Util.Environment.Installation.GetVisXpertPath(), "bin");

            //Then we can add this directory to the Search Path of this process
            NativeMethods.SetDefaultDllDirectories(NativeMethods.LOAD_LIBRARY_SEARCH_DEFAULT_DIRS);
            NativeMethods.AddDllDirectory(VxPath);
        }

        /// <summary>
        /// This must be in an separate function in order to work. 
        /// Here we can resolve any Managed Assemblies. In this example we implemented
        /// two Assembly Resolvers. One for VisXpert assemblies, and one for BatchXpert Assemblies
        /// If an Assembly is found, it will be loaded, otherwise the default search Path will be followed
        /// </summary>
        private static void SetUpAssemblyResolve()
        {
            AppDomain.CurrentDomain.AssemblyResolve+=(a, b) =>
            {
                try
                {
                    Assembly asm = null; asm=VisXpertDirectoryAssemblyResolver.Resolve(a, b);
                    if (asm != null) return asm;

                    asm=BatchXpertDirectoryAssemblyResolver.Resolve(a, b);
                    if (asm != null) return asm; return asm;
                }
                catch (Exception)
                {
                    return null;
                }
            };
        }

        private partial class NativeMethods
        {
            public const uint LOAD_LIBRARY_SEARCH_DEFAULT_DIRS = 0x1000U;

            [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
            public static extern bool SetDefaultDllDirectories(uint DirectoryFlags);

            [DllImport("kernel32.dll", CharSet = CharSet.Unicode, SetLastError = true)]
            public static extern int AddDllDirectory(string NewDirectory);
        }
    }
}
